@@ -1,7 +1,9 @@
 __author__ = 'calvin'
 
+import sys
 import traceback
 import os
+import re
 import glob
 import datetime
 import shutil
@@ -9,15 +11,17 @@ import smtplib
 import time
 import logging
 import ftplib
-import jinja2
 import ConfigParser
-
 from threading import Thread
 from email.mime.multipart import MIMEMultipart
 from email.mime.base import MIMEBase
 from email.mime.text import MIMEText
 from email import encoders
 
+import jinja2
+
+from tools import analyze_traceback
+
 
 class CrashReporter(object):
     """
@@ -27,13 +31,23 @@ class CrashReporter(object):
     If a crash report fails to upload, the report is saved locally to the `report_dir` directory. The next time the
     CrashReporter starts up, it will attempt to upload all offline reports every `check_interval` seconds. After a
     successful upload the offline reports are deleted. A maximum of `offline_report_limit` reports are saved at any
-    time. Reports are named crashreport.1, crashreport.2, crashreport.3 and so on. The most recent report is always
-    crashreport.1.
+    time. Reports are named crashreport01, crashreport02, crashreport03 and so on. The most recent report is always
+    crashreport01.
+
+    Report Customizing Attributes:
+
+    application_name: Application name as a string to be included in the report
+    application_version: Application version as a string to be included in the report
+    user_identifier: User identifier as a string to add to the report
+    offline_report_limit: Maximum number of offline reports to save.
+    max_string_length: Maximum string length for values returned in variable inspection. This prevents reports which
+                       contain array data from becoming too large.
+    inspection_level: The number of traceback objects (from most recent) to inspect for source code, local variables etc
 
     :param report_dir: Directory to save offline reports.
-    :param check_interval: How often the to attempt to send offline reports
+    :param watcher: Enable a thread that periodically checks for any stored offline reports and attempts to send them.
+    :param check_interval: How often the watcher will attempt to send offline reports.
     :param logger: Optional logger to use.
-    :param offline_report_limit: Number of offline reports to save.
     :param config: Path to configuration file that defines the arguments to setup_smtp and setup_ftp. The file has the
                    format of a ConfigParser file with sections [SMTP] and [FTP]
     :param html: Create HTML reports (True) or plain text (False).
@@ -41,40 +55,43 @@ class CrashReporter(object):
     """
     _report_name = "crashreport%02d"
     html_template = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'crashreport.html')
-    ''' Application name as a string to be included in the report'''
+    active = False
     application_name = None
-    ''' Application version as a string to be included in the report'''
     application_version = None
-    ''' The number of source code lines to include before and after the error occurs'''
-    source_code_line_limit = (25, 25)
-    ''' Number of offline reports to save.'''
+    user_identifier = None
     offline_report_limit = 10
+    max_string_length = 1000
+    inspection_level = 1
+    obj_ref_regex = re.compile("[A-z]+[0-9]*\.(?:[A-z]+[0-9]*\.?)+(?!\')")
 
-    def __init__(self, report_dir=None, html=False, check_interval=5*60, config='', logger=None):
+    def __init__(self, report_dir=None, html=True, config='', logger=None, activate=True,
+                 watcher=True, check_interval=5*60):
         self.html = html
-        self._enabled = True
         self.logger = logger if logger else logging.getLogger(__name__)
         # Setup the directory used to store offline crash reports
         self.report_dir = report_dir
         self.check_interval = check_interval
+        self.watcher_enabled = watcher
         self._watcher = None
-        self._watcher_enabled = False
-
+        self._watcher_running = False
+        self.etype = None
+        self.evalue = None
+        self.tb = None
+        self.tb_info = None
+        self._excepthook = None
+        # Load the configuration from a file if specified
         if os.path.isfile(config):
             self.load_configuration(config)
         else:
             self._smtp = None
             self._ftp = None
-
-        if report_dir:
-            if os.path.exists(report_dir):
-                self.start_watcher()
-            else:
-                os.makedirs(report_dir)
+        if activate:
+            self.enable()
 
     def setup_smtp(self, host, port, user, passwd, recipients, **kwargs):
         """
         Set up the crash reporter to send reports via email using SMTP
+
         :param host: SMTP host
         :param port: SMTP port
         :param user: sender email address
@@ -83,10 +100,12 @@ class CrashReporter(object):
         """
         self._smtp = kwargs
         self._smtp.update({'host': host, 'port': port, 'user': user, 'passwd': passwd, 'recipients': recipients})
+        self._smtp['from'] = kwargs.get('from', user)
 
     def setup_ftp(self, host, user, passwd, path, port=21, acct='', timeout=5, **kwargs):
         """
         Set up the crash reporter to upload reports via FTP.
+
         :param host: FTP host
         :param user: FTP user
         :param passwd: FTP password for user
@@ -103,26 +122,45 @@ class CrashReporter(object):
         """
         Enable the crash reporter. CrashReporter is defaulted to be enabled on creation.
         """
-        self.logger.info('CrashReporter: Enabled')
-        self._enabled = True
+        if not CrashReporter.active:
+            CrashReporter.active = True
+            # Store this function so we can set it back if the CrashReporter is deactivated
+            self._excepthook = sys.excepthook
+            sys.excepthook = self.exception_handler
+            self.logger.info('CrashReporter: Enabled')
+            if self.report_dir:
+                if os.path.exists(self.report_dir):
+                    if self._get_offline_reports():
+                        # First attempt to send the reports, if that fails then start the watcher
+                        if any(self.submit_offline_reports(smtp=True, ftp=True)):
+                            self.delete_offline_reports()
+                        elif self.watcher_enabled:
+                            self.start_watcher()
+                else:
+                    os.makedirs(self.report_dir)
 
     def disable(self):
         """
         Disable the crash reporter. No reports will be sent or saved.
         """
-        self.logger.info('CrashReporter: Disabled')
-        self._enabled = False
-        self.stop_watcher()
+        if CrashReporter.active:
+            CrashReporter.active = False
+            # Restore the original excepthook
+            sys.excepthook = self._excepthook
+            self.stop_watcher()
+            self.logger.info('CrashReporter: Disabled')
 
     def start_watcher(self):
         """
         Start the watcher that periodically checks for offline reports and attempts to upload them.
         """
-        if self._get_offline_reports():
+        if self._watcher and self._watcher.is_alive:
+            self._watcher_running = True
+        else:
             self.logger.info('CrashReporter: Starting watcher.')
             self._watcher = Thread(target=self._watcher_thread, name='offline_reporter')
             self._watcher.setDaemon(True)
-            self._watcher_enabled = True
+            self._watcher_running = True
             self._watcher.start()
 
     def stop_watcher(self):
@@ -130,31 +168,43 @@ class CrashReporter(object):
         Stop the watcher thread that tries to send offline reports.
         """
         if self._watcher:
-            self._watcher_enabled = False
+            self._watcher_running = False
             self.logger.info('CrashReporter: Stopping watcher.')
 
-    def __enter__(self):
-        self.enable()
+    def exception_handler(self, etype, evalue, tb):
+        """
+        Catches crashes/ un-caught exceptions. Creates and attempts to upload the crash reports. Calls the default
+        exception handler (sys.__except_hook__) upon completion.
 
-    def __exit__(self, etype, evalue, tb):
-        if self._enabled:
+        :param etype: Exception type
+        :param evalue: Exception value
+        :param tb: Traceback
+        :return:
+        """
+        if CrashReporter.active:
             if etype:
-                self._etype = etype
-                self._evalue = evalue
-                self._tb = tb
+                self.etype = etype
+                self.evalue = evalue
+                self.tb = tb
+                self.tb_info = analyze_traceback(tb)
+                # Save the offline report. If the upload of the report is successful, then delete the report.
+                report_path = self._save_report()
                 great_success = False
                 if self._smtp is not None:
                     # Send the report via email
-                    great_success |= self._sendmail(self.subject(), self.body(), self.attachments(), html=self.html)
+                    with open(report_path, 'r') as _cr:
+                        body = _cr.read()
+                    great_success |= self._sendmail(self.subject(), body, self.attachments(), html=self.html)
                 if self._ftp is not None:
                     # Send the report via FTP
-                    great_success |= self._ftp_submit()
+                    great_success |= self._ftp_submit(report_path)
+                if great_success:  # Very nice..
+                    os.remove(report_path)
+            else:
+                self.logger.info('CrashReporter: No crashes detected.')
 
-                # If both FTP and email sending fails, save the report
-                if not great_success:
-                    self._save_report()
-        else:
-            self.logger.info('CrashReporter: No crashes detected.')
+        # Call the default exception hook
+        sys.__excepthook__(etype, evalue, tb)
 
     def load_configuration(self, config):
         cfg = ConfigParser.ConfigParser()
@@ -185,68 +235,52 @@ class CrashReporter(object):
         else:
             return 'Crash Report'
 
-    def body(self):
+    def html_body(self):
         """
         Return a string to be used as the email body. Can be html if html is turned on.
         """
-        # Get the last traceback
-        tb_last = self._tb.tb_next
-        if tb_last is None:
-            tb_last = self._tb
-        else:
-            while tb_last.tb_next is not None:
-                tb_last = tb_last.tb_next
+        dt = datetime.datetime.now()
+        error = traceback.format_exception_only(self.etype, self.evalue)[0].strip()
+        fields = {'date': dt.strftime('%d %B %Y'),
+                  'time': dt.strftime('%I:%M %p'),
+                  'traceback': self.tb_info,
+                  'error': error,
+                  'app_name': self.application_name,
+                  'app_version': self.application_version,
+                  'user': self.user_identifier
+                  }
 
-        if self.html:
-            dt = datetime.datetime.now()
-            tb = [dict(zip(('file', 'line', 'module', 'code'),  t)) for t in traceback.extract_tb(self._tb)]
-            error = traceback.format_exception_only(self._etype, self._evalue)[0]
+        with open(self.html_template, 'r') as _f:
+            template = jinja2.Template(_f.read())
+        html_body = template.render(**fields)
+        return html_body
 
-            scope_lines = self._get_source()
-            scope_locals = self._get_locals(tb_last)
+    def raw_body(self):
 
-            fields = {'date': dt.strftime('%d %B %Y'),
-                      'time': dt.strftime('%I:%M %p'),
-                      'traceback': tb,
-                      'error': error,
-                      'localvars': scope_locals,
-                      'app_name': self.application_name,
-                      'app_version': self.application_version,
-                      'source_code': scope_lines
-                      }
-
-            with open(self.html_template, 'r') as _f:
-                template = jinja2.Template(_f.read())
-            html_body = template.render(**fields)
-            with open('report.html', 'w') as _g:
-                _g.write(html_body)
-            return html_body
-
-        else:
-
-            body = datetime.datetime.now().strftime('%d %B %Y, %I:%M %p\n')
-            body += '\n'.join(traceback.format_exception(self._etype, self._evalue, self._tb))
-            body += '\n'
-            # Print the source code in the local scope of the error
-            body += 'Source Code:\n\n'
-            scope_lines = self._get_source(tb_last)
-            for ln, indent, line in scope_lines:
-                body += "{ln}.{indent}{line}".format(ln=ln, indent=int(indent / 30. * 4) * ' ', line=line)
-            body += '\n'
-            # Print a table of local variables
-            scope_locals = self._get_locals(tb_last)
-            limit = 25
-            fmt = "{name:<25s}{value:<25s}\n"
-            body += '-' * 90 + '\n'
-            body += fmt.format(name='Variable', value='Value')
-            body += '-' * 90 + '\n'
-            count = 0
-            for name, value in scope_locals:
-                body += fmt.format(name=name, value=value.__repr__())
-                count += 1
-                if count > limit:
-                    break
-            return body
+        body = datetime.datetime.now().strftime('%d %B %Y, %I:%M %p\n')
+        body += '\n'.join(traceback.format_exception(self.etype, self.evalue, self.tb))
+        body += '\n'
+        # Print the source code in the local scope of the error
+        body += 'Source Code:\n\n'
+        scope_lines = self.tb_info[-1]['source']
+        for ln, line in scope_lines:
+            body += "{ln}.{line}".format(ln=ln, line=line)
+        body += '\nLocal Variables in the scope of {}\n'.format(self.tb_info[-1]['module'])
+        # Print a table of local variables
+        fmt = "{name:<25s}{value:<25s}\n"
+        body += '-' * 90 + '\n'
+        body += fmt.format(name='Variable', value='Value')
+        body += '-' * 90 + '\n'
+        for name, value in self.tb_info[-1]['local_vars']:
+            body += fmt.format(name=name, value=value)
+        body += '\nObject Inspection in the scope of {}\n'.format(self.tb_info[-1]['module'])
+        # Print a table of object attribute references
+        body += '-' * 90 + '\n'
+        body += fmt.format(name='Variable', value='Value')
+        body += '-' * 90 + '\n'
+        for name, value in self.tb_info[-1]['object_vars']:
+            body += fmt.format(name=name, value=value)
+        return body
 
     def attachments(self):
         """
@@ -260,9 +294,32 @@ class CrashReporter(object):
         Delete all stored offline reports
         """
         for report in self._get_offline_reports():
-            os.remove(report)
+            try:
+                os.remove(report)
+            except OSError as e:
+                logging.error(e)
 
-    def _ftp_submit(self):
+    def submit_offline_reports(self, smtp=True, ftp=True):
+        """
+        Submit offline reports using the enabled methods (SMTP and/or FTP)
+        Returns a tuple of booleans specifying which methods passed (smtp_success, ftp_success)
+        """
+        ftp_success = smtp_success = False
+
+        if smtp and self._smtp is not None:
+            try:
+                smtp_success = self._smtp_send_offline_reports()
+            except Exception as e:
+                logging.error(e)
+        if ftp and self._ftp is not None:
+            try:
+                ftp_success = self._ftp_send_offline_reports()
+            except Exception as e:
+                logging.error(e)
+
+        return smtp_success, ftp_success
+
+    def _ftp_submit(self, path):
         """
         Upload the database to the FTP server. Only submit new information contained in the partial database.
         Merge the partial database back into master after a successful upload.
@@ -276,17 +333,12 @@ class CrashReporter(object):
             self.logger.error(e)
             self.stop_watcher()
             return False
-
-        tmp_file = 'ftp_report'
-        tmp_path = self._write_report(tmp_file)
-
+        extension = os.path.splitext(path)[1]
         ftp.cwd(info['path'])
-        with open(tmp_path, 'rb') as _f:
-            ext = '.html' if self.html else '.txt'
-            new_filename = self._report_name % (len(ftp.nlst()) + 1) + ext
+        with open(path, 'rb') as _f:
+            new_filename = self._report_name % (len(ftp.nlst()) + 1) + extension
             ftp.storlines('STOR %s' % new_filename, _f)
-            os.remove(tmp_path)
-            self.logger.info('CrashReporter: Submission to %s successful.' % info['host'])
+            self.logger.info('CrashReporter: FTP submission to %s successful.' % info['host'])
             return True
 
     def _ftp_send_offline_reports(self):
@@ -309,7 +361,7 @@ class CrashReporter(object):
                 ext = os.path.splitext(report)[1]
                 new_filename = self._report_name % (len(ftp.nlst()) + 1) + ext
                 ftp.storlines('STOR %s' % new_filename, _f)
-        self.logger.info('CrashReporter: Submission to %s successful.' % info['host'])
+                self.logger.info('CrashReporter: Submission of %s to %s successful.' % (new_filename, info['host']))
         return True
 
     def _sendmail(self, subject, body, attachments=None, html=False):
@@ -319,7 +371,7 @@ class CrashReporter(object):
             msg['To'] = ', '.join(smtp['recipients'])
         else:
             msg['To'] = smtp['recipients']
-        msg['From'] = smtp['user']
+        msg['From'] = smtp['from']
         msg['Subject'] = subject
 
         # Add the body of the message
@@ -344,7 +396,7 @@ class CrashReporter(object):
             ms.starttls()
             ms.ehlo()
             ms.login(smtp['user'], smtp['passwd'])
-            ms.sendmail(smtp['user'], smtp['recipients'], msg.as_string())
+            ms.sendmail(smtp['from'], smtp['recipients'], msg.as_string())
             ms.close()
         except Exception as e:
             self.logger.error('CrashReporter: %s' % e)
@@ -369,42 +421,10 @@ class CrashReporter(object):
                 self.logger.info('CrashReporter: Offline reports sent.')
             return great_success
 
-    def _get_source(self):
-        scope_lines = []
-        _file, line_num, func, text = traceback.extract_tb(self._tb)[-1]
-        start = line_num - self.source_code_line_limit[0]
-        end = line_num + self.source_code_line_limit[1]
-        with open(_file, 'r') as _f:
-            for c, l in enumerate(_f):
-                if start <= c + 1 <= end:
-                    scope_lines.append((c+1, 30 * (l.count('    ')-1), l.replace('    ', '')))
-        return scope_lines
-
-    def _get_locals(self, tb):
-        if 'self' in tb.tb_frame.f_locals:
-            _locals = [('self', tb.tb_frame.f_locals['self'].__repr__())]
-        else:
-            _locals = []
-        for k, v in tb.tb_frame.f_locals.iteritems():
-            if k == 'self':
-                continue
-            try:
-                _locals.append((k, v.__repr__()))
-            except TypeError:
-                pass
-        return _locals
-
-    def _write_report(self, path):
-        # Write a new report
-        path = os.path.splitext(path)[0] + ('.html' if self.html else '.txt')
-        with open(path, 'w') as _f:
-            _f.write(self.body())
-        return path
-
     def _save_report(self):
         """
-        Save the crash report to a file. Keeping the last 5 files in a cyclical FIFO buffer.
-        The newest crash report is 01
+        Save the crash report to a file. Keeping the last `offline_report_limit` files in a cyclical FIFO buffer.
+        The newest crash report always named is 01
         """
         offline_reports = self._get_offline_reports()
         if offline_reports:
@@ -419,8 +439,12 @@ class CrashReporter(object):
             if len(offline_reports) >= self.offline_report_limit:
                 oldest = glob.glob(os.path.join(self.report_dir, self._report_name % (self.offline_report_limit+1) + '*'))[0]
                 os.remove(oldest)
-        new_report_path = os.path.join(self.report_dir, self._report_name % 1)
-        self._write_report(new_report_path)
+        new_report_path = os.path.join(self.report_dir, self._report_name % 1 + ('.html' if self.html else '.txt'))
+        # Write a new report
+        with open(new_report_path, 'w') as _f:
+            _f.write(self.html_body() if self.html else self.raw_body())
+
+        return new_report_path
 
     def _get_offline_reports(self):
         return sorted(glob.glob(os.path.join(self.report_dir, "crashreport*")))
@@ -432,15 +456,11 @@ class CrashReporter(object):
         great_success = False
         while not great_success:
             time.sleep(self.check_interval)
-            if not self._watcher_enabled:
+            if not self._watcher_running:
                 break
             self.logger.info('CrashReporter: Attempting to send offline reports.')
-            if self._smtp is not None:
-                great_success |= self._smtp_send_offline_reports()
-            if self._ftp is not None:
-                # Send the report via FTP
-                great_success |= self._ftp_send_offline_reports()
-
+            great_success |= any(self.submit_offline_reports(smtp=True, ftp=True))
         if great_success:
             self.delete_offline_reports()
-        self.logger.info('CrashReporter: Watcher stopped.')
\ No newline at end of file
+        self._watcher = None
+        self.logger.info('CrashReporter: Watcher stopped.')
